#include <stdio.h>

  /* Broj particija p(n) - Eulerova rekurzivna formula:
     p(n) = broj rastava zadanog prirodnog broja n
     u sumu nepadajucih (prirodnih) pribrojnika.
     Argument n smije biti cijeli broj.
     Dogovorno je p(0) = 1 i p(n) = 0, za n < 0.

     Rekurzivna funkcija za p (po Eulerovoj formuli).
     Napomena: broj rekurzivnih poziva strelovito RASTE,
     puno BRZE nego kod Fibonaccijevih brojeva!

     Umjesto citanja broja n, ispisuje broj particija p(n),
     za sve brojeve n = 0, ..., 40.
     Ispisuje i broj rekurzivnih poziva funkcije p.
     Za n = 40, broj poziva vise NIJE prikaziv u tipu unsigned int!
  */

    /* Globalni brojac rekuzivnih poziva funkcije. */

unsigned int broj_poziva;


    /* Eulerova formula za broj particija p(n), za n >= 1:
         p(n) = suma_(m = 1)^(beskonacno)
                  (-1)^(m - 1) * ( p(n1_m) + p(n2_m) ),
       gdje su argumenti na desnoj strani (u sumi)
         n1_m = n - m * (3m - 1) / 2,
         n2_m = n - m * (3m + 1) / 2 = n1_m - m,
       uz p(0) = 1 i p(n) = 0, za n < 0.

       Implementacija: Uociti da suma ide najdalje do m = n.
       Nadalje, za argumente vrijedi n1_m > n2_m = n1_m - m.
       U uzlaznoj petlji po m, prvo racunamo veci argument n1_m.
       Kad dobijemo da je n1_m < 0, prekidamo petlju po m.

       Umjesto gornje izravne formule za n1_m, koristimo vezu
       izmedju dva susjedna clana (po m):
         n1_m = n1_(m - 1) - 3*m + 2, za m >= 1,
       uz pocetak n1_0 = n. To malo ubrzava racunanje. */


    /* Rekurzivna funkcija za broj particija p(n). */

    /* Napomena: Prvih nekoliko clanova u Eulerovoj formuli je
         p(n) = p(n - 1) + p(n - 2) - p(n - 5) - p(n - 7) ...
       tako da je broj rekurzivnih poziva VECI nego kod
       Fibonaccijevih brojeva! */

int p(int n)
{
    int broj = 0, m, n1_m, n2_m, pribr;

        /* Povecaj globalni brojac poziva funkcije. */
    ++broj_poziva;

        /* Provjera kraja rekurzije, za n <= 0. */
    if (n < 0) return 0;
    if (n == 0) return 1;

        /* Pocetni veci argument n1_0 = n, za m = 0. */
    n1_m = n;

        /* Suma iz Eulerove formule - uzlazna petlja po m.
           Vrijednosti funkcije p za argumente n1_m, n2_m
           (oba su < n) racunamo rekurzivnim pozivima funkcije p.
           Pazimo da su argumenti >= 0 (nema nepotrebnih poziva). */

    for (m = 1; m <= n; ++m) {

            /* Prvo racunamo veci argument n1_m iz prethodnog. */
        n1_m = n1_m - 3*m + 2;

            /* Kad dobijemo n1_m < 0, prekidamo petlju po m. */
        if (n1_m < 0) break;

            /* Rekurzivni poziv za p(n1_m). */
        pribr = p(n1_m);

            /* Manji argument n2_m iz n1_m. */
        n2_m = n1_m - m;

            /* Rekurzivni poziv za p(n2_m), ako je n2_m >= 0. */
        if (n2_m >= 0)
            pribr += p(n2_m);

            /* Provjera predznaka (parnosti od m) - ovisno o tome,
               broju treba dodati ili oduzeti pribrojnik pribr. */
        if (m % 2 == 1)    /* (m - 1) je paran. */
            broj += pribr;
        else
            broj -= pribr;
    }

    return broj;
}

    /* Glavni program za tablicu broja particija. */

int main(void)
{
    int n;

        /* Zaglavlje tablice. */
    printf("\n");
    printf("     Broj particija      Broj poziva\n");
    printf(" -------------------------------------\n");

    for (n = 0; n <= 40; ++n ) {

            /* Inicijalizacija globalnog brojaca poziva. */
        broj_poziva = 0u;

            /* Redak tablice. */
        printf("   p(%3d) = %10d    %10u\n",
               n, p(n), broj_poziva);
    }

    return 0;
}
