#include <stdio.h>
#include <stdlib.h>

  /* Preuredjenje vezane liste - prvo parni, onda neparni.
     CUVA relativni poredak parnih, odnosno, neparnih elemenata.

     Zbog CUVANJA poretka, algoritam moze napredovati po BLOKOVIMA.
     BLOK = najdulja podlista uzastopnih elemenata iste vrste (parnosti).

     Algoritam b2:
     - Rastav polazne liste s jednom zasebnom listom, spajanje na kraju.
     - Zasebna lista NIJE stalno uzemljena.
     - Zavrsna lista se uzemljava na kraju, kod spajanja.
     - Napreduje po blokovima elemenata iste vrste (parnosti).
     Vrlo EFIKASAN, jer izbjegava nepotrebno azuriranje pokazivaca
     (nema promjena veza unutar jednog bloka).
  */

/* ======================================================================== */

    /* Tip za pokazivac na element liste. */

typedef struct _element *lista;

    /* Tip za element liste. */

typedef struct _element {
    int broj;            /* Sadrzaj je broj. */
    lista sljed;         /* Pokazivac na sljedeci element u listi. */
} element;

/* ======================================================================== */

    /* Preuredjuje listu tako da na pocetku liste budu svi parni,
       a zatim dolaze svi neparni elementi iz polazne liste.
       Relativni poredak parnih, odnosno, neparnih elemenata
       je ISTI kao u polaznoj listi. */

    /* Uvod u algoritam:
         Lista sadrzi dvije vrste elemenata (ovdje su to parni i neparni).
       BLOK = najdulja podlista uzastopnih elemenata iste vrste (parnosti).
       Preuredjena lista na pocetku ima blok jedne vrste elemenata (parni),
       a zatim blok druge vrste (neparni).

         Ideja algoritma: rastavi polaznu listu u dva bloka, svaki blok
       ima elemente svoje vrste, a zatim ih spoji.

         Zbog CUVANJA relativnog poretka elemenata iste vrste, algoritam
       moze napredovati po BLOKOVIMA, jer nema potrebe za promjenom veza
       UNUTAR jednog bloka (relativni poredak elemenata u jednom bloku
       ostaje isti).

       Bitno POJEDNOSTAVLJENJE algoritma ("zrno soli"):
         Kod rastava liste, NE treba fiksirati znacenje pojedine vrste
       elemenata, tj. NIJE bitno da prva vrsta budu parni, a druga
       neparni elementi. Dozvoljeno je i obratno! Puno lakse je ovako:
       - Prva vrsta odgovara prvom bloku u polaznoj listi, a
         druga vrsta odgovara drugom bloku u polaznoj listi.
       Ako ti blokovi nisu prazni, onda MORAJU biti razlicitih vrsta.
       Tek kod spajanja, treba paziti koja vrsta ide sprijeda.
       Zato liste, koje sadrze elemente odgovarajuce vrste, oznacavamo
       indeksima 1 i 2 (prva i druga vrsta), a ne parni i neparni.
    */

    /* Algoritam b2:
         Rastavi polaznu listu tako da jedna vrsta elemenata ostaje
       u listi (na pocetku, u prednjem = sredjenom dijelu), a druga
       vrsta elemenata se izbacuje i prebacuje u (jednu) zasebnu listu.
       Na kraju, spoji te dvije liste (konkatencija).
       Zasebna lista NIJE stalno uzemljena (za povecanje efikasnosti).

         Drugim rijecima, umjesto gradjenja dvije zasebne liste, jedna
       od njih se gradi u prednjem (sredjenom) dijelu "radne" liste.
       IZBACIVANJE blokova druge vrste (uz korektno vezanje preostalih
       blokova prve vrste) osigurava CUVANJE relativnog poretka prvih!

         Za CUVANJE relativnog poretka, radimo ubacivanje blokova na
       KRAJ odgovarajuce liste. Zato pamtimo prvi i zadnji element u
       obje liste. Dodatno, to bitno ubrzava spajanje (konkatenaciju)
       te dvije liste na kraju algoritma, jer ne treba traziti zadnjeg
       u jednoj (prednjoj) od njih.
       Zato sto se jedna lista gradi u radnoj listi, zadnji sredjeni
       element je upravo "prethodnik" elementa kojeg obradjujemo,
       tj. prvi_1 <=> prvi, zadnji_1 <=> preth.

         Dakle, algoritam se svodi na izbacivanje blokova druge vrste
       iz polazne liste i gradjenje zasebne liste drugih, dodavanjem
       na kraj.

         Unutar petlje za obradu, sljedbenik zadnjeg u zasebnoj listi
       (neparni) se NE postavlja na NULL. Korektna vrijednost se
       postavlja tek kod spajanja.
    */

lista preuredi_par_nepar_b2(lista prvi)
{
        /* Inicijalizacija liste elemenata prve vrste. Ona ostaje u
           polaznoj listi. Zato sto prva vrsta odgovara prvom bloku
           u polaznoj listi, uvijek vrijedi prvi_1 = prvi. Dovoljno
           je pamtiti pamtiti pokazivac na zadnji element prve vrste.
           NE treba ga inicijalizirati na NULL.
           Ujedno, to je i pokazivac na element ispred pocetka sljedeceg
           bloka kojeg treba obraditi, tj. zadnji_1 = preth. */
    lista zadnji_1;

        /* Inicijalizacija liste elemenata druge vrste na praznu listu.
           Za listu pamtimo pokazivace na prvi i zadnji element.
           Pokazivac na prvi element inicijaliziramo na NULL.
           Pokazivac na zadnji element NE treba inicijalizirati na NULL. */
    lista prvi_2 = NULL, zadnji_2;

        /* Ostatak pri dijeljenju s 2, za elemente prve vrste.
           On definira prvu vrstu - parni ili neparni. Ostatak za
           drugu vrstu ne treba pamtiti, jer imamo samo dvije vrste
           elemenata (ostatak_2 = 1 - ostatak_1). */
    int ostatak_1;

        /* Pokazivaci na radni blok elemenata (iste vrste).
           Pamtimo pokazivace na prvi i zadnji element u bloku. */
    lista prvi_b, zadnji_b;

        /* Ostatak pri dijeljenju s 2, za elemente radnog bloka.
           On odredjuje vrstu radnog bloka. */
    int ostatak_b;

        /* Pokazivac na radni element, kod prolaza po listi.
           Inicijaliziramo ga na prvi. */
    lista radni = prvi;


        /* Nadji i obradi PRVI blok (= prva vrsta), na pocetku liste. */

    if (prvi != NULL) {    /* <=> radni != NULL. */

            /* Prvi blok sigurno nije prazan. Njegov pocetak je prvi.
               Inicijaliziraj zadnji element u bloku na prvi. Zapamti
               ostatak pri dijeljenju prvog elementa s 2 (vrsta bloka).
               To definira prvu vrstu i sluzi za nalazenje kraja bloka. */
        zadnji_1 = prvi;
        ostatak_1 = prvi->broj % 2;    /* Definira prvu vrstu. */

            /* Sljedeci radni element - za trazenje kraja bloka. */
        radni = prvi->sljed;

            /* Nadji kraj bloka. Micemo radni, sve dok ne naidjemo na
               NULL ili na element razlicite vrste (razlicit ostatak).
               Koristimo skraceno racunanje logickih izraza.
               Zadnji element u bloku je onaj ISPRED radnog. */
        while (radni != NULL && radni->broj % 2 == ostatak_1) {
            zadnji_1 = radni;
            radni = radni->sljed;
        }

            /* Prvi blok ostaje na pocetku liste - nema izbacivanja
               (tako smo izabrali prvu vrstu - da ostaje u listi).
               Ne treba uzemljiti zadnji_1. */
    }
    else    /* prvi == NULL <=> radni == NULL. */

            /* Prvi blok je prazan => cijela preuredjena lista je prazna.
               Odmah vrati NULL. */
        return NULL;


        /* Nadji i obradi DRUGI blok (= druga vrsta), na pocetku liste. */

    if (radni != NULL) {

            /* Drugi blok sigurno nije prazan. Zapamti njegov pocetak i
               inicijaliziraj zadnji element u bloku na prvi. Zapamti
               ostatak pri dijeljenju prvog elementa s 2 (vrsta bloka).
               Taj ostatak sluzi samo za nalazenje kraja bloka (znamo da
               mora biti razlicit od prvog, tj. blok je druge vrste). */
        prvi_2 = radni;
        zadnji_2 = prvi_2;
        ostatak_b = prvi_2->broj % 2;

            /* Sljedeci radni element - za trazenje kraja bloka. */
        radni = radni->sljed;

            /* Nadji kraj bloka. Micemo radni, sve dok ne naidjemo na
               NULL ili na element razlicite vrste (razlicit ostatak).
               Koristimo skraceno racunanje logickih izraza.
               Zadnji element u bloku je onaj ISPRED radnog. */
        while (radni != NULL && radni->broj % 2 == ostatak_b) {
            zadnji_2 = radni;
            radni = radni->sljed;
        }

            /* Izbaci drugi blok iz polazne liste - neobradjeni dio
               polazne liste (na kojeg pokazuje radni) treba spojiti
               iza zadnji_1. */
        zadnji_1->sljed = radni;

            /* Ne treba uzemljiti zadnji_2. */
    }
    else    /* radni == NULL. */

            /* Drugi blok je prazan => cijela lista ima samo prvi blok
               (tj. polazna lista je vec uredjena). Uzemljenje prvog
               bloka (zadnji_1->sljed = NULL;) nije potrebno, jer je
               vec uzemljen (cijela lista). Samo vrati pocetak. */
        return prvi;


        /* Na ovom mjestu, prva i druga lista vec imaju neprazne pocetne
           blokove, pa vise ne treba paziti na pocetke - samo dodajemo
           blokove na kraj odgovarajuce liste (prema vrsti bloka).

           Napomena: U nastavku obrade, NE treba paziti na to da vrste
           blokova "alterniraju" - parni, pa neparni (ili obratno).
           Naprosto obradjujemo jedan po jedan blok i dodajemo ga u
           odgovarajucu listu, prema vrsti (ostatku u tom bloku)! */

        /* Petlja za obradu preostalih blokova u polaznoj listi.
           Napreduje za po jedan blok elemenata (iste vrste).
           U trenutku provjere uvjeta petlje, pokazivac radni pokazuje
           na pocetak neobradjenog dijela liste = pocetak sljedeceg bloka
           (ako postoji). */

    while (radni != NULL) {

            /* Sljedeci blok sigurno nije prazan. Zapamti njegov pocetak i
               inicijaliziraj zadnji element u bloku na prvi. Zapamti
               ostatak pri dijeljenju prvog elementa s 2 (vrsta bloka).
               Taj ostatak sluzi za nalazenje kraja bloka i kasniju
               obradu bloka (prema vrsti). */
        prvi_b = radni;
        zadnji_b = prvi_b;
        ostatak_b = prvi_b->broj % 2;

            /* Sljedeci radni element - za trazenje kraja bloka. */
        radni = radni->sljed;

            /* Nadji kraj bloka. Micemo radni, sve dok ne naidjemo na
               NULL ili na element razlicite vrste (razlicit ostatak).
               Koristimo skraceno racunanje logickih izraza.
               Zadnji element u bloku je onaj ISPRED radnog. */
        while (radni != NULL && radni->broj % 2 == ostatak_b) {
            zadnji_b = radni;
            radni = radni->sljed;
        }

            /* Ne treba uzemljiti zadnji_b. */

            /* Dodaj nadjeni blok na kraj odgovarajuce liste i postavi
               novi zadnji u toj listi (ne treba ga uzemljiti). */
        if (ostatak_b == ostatak_1) {
            zadnji_1->sljed = prvi_b;
            zadnji_1 = zadnji_b;
        }
        else {
            zadnji_2->sljed = prvi_b;
            zadnji_2 = zadnji_b;
        }
    }

        /* Spajanje (konkatenacija) prve i druge liste (obje su neprazne).
           U spojenoj listi, parni moraju niti ispred neparnih.
           Liste NISU uzemljene, pa treba paziti na to.
           - Spojimo neparnu listu na parnu.
           - Uzemljimo neparnu listu (jer ona ide na kraj spojene).
           - Vratimo pokazivac na pocetak spojene liste = pocetak parne. */

    if (ostatak_1 == 0) {    /* Prva je parna, druga je neparna. */
        zadnji_1->sljed = prvi_2;
        zadnji_2->sljed = NULL;
        return prvi;
    }
    else {    /* Druga je parna, prva je neparna. */
        zadnji_2->sljed = prvi;
        zadnji_1->sljed = NULL;
        return prvi_2;
    }
}

/* ======================================================================== */

    /* Ispod su funkcije za osnovne operacije s listama. */

/* ======================================================================== */

lista kreiraj_straga(lista prvi, lista *p_zadnji, int broj)
{
        /* Vraca zadnji kroz varijabilni argument - pokazivac p_zadnji. */

    lista zadnji = *p_zadnji;
    lista novi = NULL;
/*
    if ((novi = (lista) malloc(sizeof(element))) == NULL)
*/
    novi = (lista) malloc(sizeof(element));
    if (novi == NULL) {
        printf("Alokacija nije uspjela.\n");
        exit(EXIT_FAILURE);    /* exit(1); */
    }

    novi->broj = broj;
    novi->sljed = NULL;

    if (prvi == NULL)
        prvi = novi;
    else
        zadnji->sljed = novi;

/* Ne treba:
    zadnji = novi; *p_zadnji = zadnji;
*/
    *p_zadnji = novi;    /* Vrati novi zadnji! */

    return prvi;
}

/* ======================================================================== */

lista obrisi_listu(lista prvi)
{
    lista pom;
/*
    while (prvi != NULL)
        prvi = obrisi_prvog(prvi);
*/
    while (prvi != NULL) {
        pom = prvi;
        prvi = prvi->sljed;
        free(pom);
    }

    return NULL;    /* <=> return prvi; */
}

/* ======================================================================== */

int broj_elemenata(lista prvi)
{
    lista pom;
    int brojac = 0;

    for (pom = prvi; pom != NULL; pom = pom->sljed)
        ++brojac;

    return brojac;
}

/* ======================================================================== */

void ispisi_listu(lista prvi)
{
    lista pom;
    int brojac = 0;

    for (pom = prvi; pom != NULL; pom = pom->sljed) {
        printf(" Element %2d, broj = %2d\n",
               ++brojac, pom->broj);
    }

    return;
}

/* ======================================================================== */

int main(void)
{
    lista prvi = NULL;    /* Prazna lista! */
    lista zadnji;
    int broj;

        /* Punjenje liste straga. */

        /* Citanje niza brojeva za listu, do broj = 0.
           To je oznaka kraj niza, ali nije clan niza. */
        /* Poredak brojeva u listi je isti kao redoslijed
           citanja, jer dodajemo na kraj liste. */

    printf("Upisite niz cijelih brojeva (0 = kraj).\n");
    do {
        printf("Sljedeci broj:\n");
        scanf("%d", &broj);
        if (broj != 0)
            prvi = kreiraj_straga(prvi, &zadnji, broj);
    } while (broj != 0);

        /* Ispis broja elemenata i cijele liste. */

    printf("\n");
    printf(" Polazna lista:\n");
    printf(" Broj elemenata u listi = %d\n",
           broj_elemenata(prvi));
    ispisi_listu(prvi);


        /* Preuredjenje liste - prvo parni, onda neparni. */

    prvi = preuredi_par_nepar_b2(prvi);

        /* Preuredjena lista:
           ispis broja elemenata i cijele liste. */

    printf("\n");
    printf(" Preuredjena lista (parni, neparni):\n");
    printf(" Broj elemenata u listi = %d\n",
           broj_elemenata(prvi));
    ispisi_listu(prvi);

        /* Obrisi cijelu listu. */

    prvi = obrisi_listu(prvi);

    return 0;
}
